"""Parsing functions."""

# Standard imports
import ast
import re
import tokenize
from io import TextIOWrapper
from pathlib import Path

# Local imports
from .classes import IgnoreFile, IgnoreLine

IGNORE_LINE_TOKEN_REGEX = r"noqa:[ ]?(?P<code>PPL[0-9]{3})"
IGNORE_FILE_TOKEN_REGEX = r"<py-printlinter disable-file (?P<code>ALL|PPL[0-9]{3})>"


def enumerate_file(folder: Path) -> list[Path]:
    """
    Enumerate all files in a folder.

    Args:
        folder: Path of the folder to enumerate

    Returns:
        All file in the given folder.
    """
    return list(folder.glob("**/*.py"))


def get_ignore_lines(file: TextIOWrapper, file_path: Path) -> list[IgnoreLine]:
    """
    Get ignored lines in a file.

    Args:
        file: File to analyze.
        file_path: Path of the given file.

    Returns:
        All ignored lines in given file.
    """
    tokens = tokenize.generate_tokens(file.readline)
    ignore_lines = []

    for token in tokens:
        toktype, tokval, start, *_ = token
        if toktype == tokenize.COMMENT:
            if match := re.search(IGNORE_LINE_TOKEN_REGEX, tokval):
                lineo, _ = start
                code = match.group("code")
                ignore_lines.append(IgnoreLine(lineo, code, file_path))

    return ignore_lines


def _no_code_before(
    tokens_before: list[tokenize.TokenInfo],
) -> bool:
    """
    Check if we have code in token before a specific token.

    Args:
        tokens_before: Token in file before the specific token.

    Returns:
        True if we have not code tokens in the given tokens, False otherwise.
    """
    for token in tokens_before:
        toktype, *_ = token
        if toktype not in [
            tokenize.COMMENT,
            tokenize.NEWLINE,
            tokenize.INDENT,
            tokenize.DEDENT,
            tokenize.ENCODING,
            tokenize.NL,
        ]:
            return False

    return True


def get_ignore_files(file: TextIOWrapper, file_path: Path) -> list[IgnoreFile] | None:
    """
    Get ignored files in a file.

    Args:
        file: File to analyze.
        file_path: Path of the given file.

    Returns:
        Ignored File in given file.
    """
    tokens = tokenize.generate_tokens(file.readline)
    tokens_before: list[tokenize.TokenInfo] = []

    for token in tokens:
        toktype, tokval, *_ = token
        if toktype == tokenize.COMMENT:
            if match := re.search(IGNORE_FILE_TOKEN_REGEX, tokval):
                if _no_code_before(tokens_before):
                    code = match.group("code")
                    return [IgnoreFile(code, file_path)]
                else:
                    return None
        tokens_before.append(token)

    return []


def parse_file(
    file_path: str,
    target_version: tuple[int, int],
) -> tuple[ast.AST, list[IgnoreLine], list[IgnoreFile] | None]:
    """
    Parse a file to get the tree and all ignore lines in this file.

    Args:
        file_path: Path of the file to parse.
        target_version: Target version for ast.

    Returns:
        The AST tree of the file and all ignored lines contains in this file.
    """
    with open(file_path, encoding="utf-8") as file:
        tree = ast.parse(
            source=file.read(),
            filename=file_path,
            feature_version=target_version,
        )
        file.seek(0)

        ignore_lines = get_ignore_lines(file, Path(file_path))
        file.seek(0)

        ignore_files = get_ignore_files(file, Path(file_path))
        file.seek(0)

    return tree, ignore_lines, ignore_files



"""Config class for the linter."""

# Standard imports
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import cast

# In python 3.11, tomllib was added,
# https://docs.python.org/3/whatsnew/3.11.html#new-modules, so that printlinter is
# compatible with versions 3.10 and 3.11 without having 2 different packages, we use the
# toml, https://pypi.org/project/toml/, for versions lower
# than python 3.11 and the tomlib of python 3.11 for versions greater than python 3.10
if sys.version_info < (3, 11):
    # Third party imports
    from toml import load as toml_load  # pragma: no cover
else:
    # Standard imports
    from tomllib import load as toml_load  # pragma: no cover

# Standard imports
from json import loads as json_load

# Third party imports
from yaml import safe_load as yaml_load

CONFIG_DEFAULT_FILES = [
    Path("printlinter.yaml"),
    Path("printlinter.yml"),
    Path("printlinter.json"),
    Path("printlinter.toml"),
    Path("pyproject.toml"),
]

MAX_MAJOR = 3
MAX_MINOR = 11


@dataclass
class Config:
    """Configuration of the linter."""

    target_version: tuple[int, int]
    "Target python version."

    ignored_files: list[str]
    "Ignored files."

    disabled_rules: list[str]
    "Disabled rules."

    def __init__(self, path: Path | None = None) -> None:
        """
        Initialize a configuration.

        We can give a path for a config file. If the path isn't given we try to load
        files in this order:
            - printlinter.yaml
            - printlinter.yml
            - printlinter.json
            - printlinter.toml
            - pyproject.toml

        Args:
            path: Optional path of a config file.
        """
        config = self._read_config(path)

        self.target_version = self._fix_target_version(config)
        self.ignored_files = cast(list[str], config.get("ignored_files", []))
        self.disabled_rules = cast(list[str], config.get("disabled_rules", []))

    def _read_config(
        self,
        path: Path | None,
    ) -> dict[str, tuple[int, int] | list[str] | str]:
        """
        Read the correct config file to produce a config dict.

        Args:
            path: Optional path of a config path.

        Returns:
            Config information.
        """
        config_files = CONFIG_DEFAULT_FILES.copy()
        if path is not None:
            if not path.exists():
                raise FileNotFoundError(
                    f"{path} does not exist. "
                    "Create it or use one of those files: "
                    f"{', '.join([str(file) for file in CONFIG_DEFAULT_FILES])}"
                )

            if path.suffix not in [".toml", ".yaml", ".yml", ".json"]:
                raise TypeError("You must give a toml, yaml or json file for config.")

            config_files.insert(0, path)

        config = None
        for config_file in config_files:
            if config_file.exists():
                config = self._load_config(config_file)
                if config is not None:
                    break

        if config is None:
            config = {
                "target_version": "3.10",
                "ignored_files": [],
                "disabled_rules": [],
            }

        return config

    def _load_config(
        self,
        path: Path,
    ) -> dict[str, tuple[int, int] | list[str] | str] | None:
        """
        Load config from a config file.

        Args:
            path: Path of the config file to load.

        Returns:
            All config from the given file.
        """
        match path.suffix:
            case ".toml":
                with open(path) as toml_file:
                    if path.stem == "pyproject":
                        loaded = toml_load(toml_file)["tool"]
                    else:
                        loaded = toml_load(toml_file)
                    return (
                        loaded.get("printlinter", None) if loaded is not None else None
                    )
            case ".yml" | ".yaml":
                with open(path) as yaml_file:
                    loaded = yaml_load(yaml_file)
                    return (
                        loaded.get("printlinter", None) if loaded is not None else None
                    )
            case ".json":
                with open(path) as json_file:
                    loaded = json_load(json_file.read())
                    return (
                        loaded.get("printlinter", None) if loaded is not None else None
                    )
            case _:  # pragma: no cover
                return None

    def _fix_target_version(
        self,
        config: dict[str, tuple[int, int] | list[str] | str],
    ) -> tuple[int, int]:
        """
        Fix `target_version` info.

        Args:
            config: Configuration values.

        Returns:
            Target version value for configuration.
        """
        target_version_from_file = str(config.get("target_version", "3.10"))
        target_version = target_version_from_file.split(".")

        err_msg = "Target version configuration must be in a format like: `3.10`"

        if len(target_version) != 2:
            raise ValueError(err_msg)

        res = []
        for num in target_version:
            if not num.isdigit():
                raise TypeError(err_msg)
            res.append(int(num))

        if not (3 <= res[0] <= MAX_MAJOR) or not (7 <= res[1] <= MAX_MINOR):
            raise ValueError(
                f"Target version must be between 3.7 and {MAX_MAJOR}.{MAX_MINOR}"
            )

        return cast(tuple[int, int], tuple(res))



"""App for typoer to interact with a user in command line."""

# Standard imports
from pathlib import Path

# Third party imports
import typer
from rich.console import Console

# First party imports
from printlinter import Config
from printlinter import __app_name__ as ppl_app_name
from printlinter import __version__ as ppl_version
from printlinter import contains_print, enumerate_file, get_not_ignore_issue, parse_file

APP = typer.Typer(help="Print linter", rich_markup_mode="rich")

CONSOLE = Console()


def version_callback(value: bool) -> None:
    """
    Get the version, display the version if `--version` was used.

    Args:
        value: Boolean to known if we display the version or not.
    """
    if value:
        CONSOLE.print(f"{ppl_app_name} v.{ppl_version}")
        raise typer.Exit()


def path_callback(path: Path) -> Path:
    """
    Check if the path exist and if it's a directory.

    Args:
        path: Path to check.

    Raises:
        FileNotFoundError: If the path does not exist.
        NotADirectoryError: If the given path is not a directory.

    Examples:
        >>> path_callback(Path("printlinter"))
        PosixPath('printlinter')

        >>> try:
        ...     path_callback(Path("azert.qwerty"))
        ... except(FileNotFoundError):
        ...     False
        False

        >>> try:
        ...     path_callback(Path("README.md"))
        ... except(TypeError):
        ...     False
        False
    """
    if not path.exists():
        raise FileNotFoundError(f"The path: {path} does not exist")
    if not path.is_dir() and path.suffix != ".py":
        raise TypeError(f"The given file {path} must be a python file (.py)")
    return path


def is_a_file(file_name: Path) -> Path | None:
    """
    Check if given path is a file.

    If the path is the default path (`Path(".")`) return None.

    Args:
        file_name: Path to check.

    Returns:
        The given path or None if given path is the default path (`Path(".")`).

    Examples:
        >>> is_a_file(Path("."))
        None

        >>> is_a_file(Path("README.md"))
        PosixPath('README.md')

        >>> try:
        ...     is_a_file(Path("printlinter"))
        ... except FileNotFoundError:
        ...     False
        False
    """
    if file_name == Path("."):
        return None
    if file_name.is_dir() or not file_name.is_file():
        raise FileNotFoundError(f"The path: {file_name} is not a file")
    return file_name


@APP.command(name="lint", help="lint the code to find print")
def lint(
    path: Path = typer.Argument(
        Path("."),
        help="Path to lint",
        show_default=False,
        callback=path_callback,
    ),
    config_file: Path = typer.Option(
        Path("."),
        help="Configuration file",
        show_default=False,
        callback=is_a_file,
    ),
) -> None:
    """
    Lint the given path or default path: `.`.

    Args:
        path: Path to lint.
        config_file: Optional config file.
    """
    warning = False
    config = Config(config_file)

    if path.is_dir():
        files_path = enumerate_file(path)
    else:
        files_path = [path]

    all_ignored_lines = []
    all_ignored_files = []
    issues = []
    for file_path in files_path:
        if file_path.absolute() in [
            Path(path).absolute() for path in config.ignored_files
        ]:
            continue

        tree, ignored_lines, ignored_files = parse_file(
            file_path.as_posix(), config.target_version
        )
        issues = contains_print(file_path, tree)
        all_ignored_lines.extend(ignored_lines)

        if ignored_files is None:
            warning = True
            CONSOLE.print(
                f"[bold][orange3]Warning ⚠️ [/orange3] {file_path}[/bold]: The ignore "
                "file comment must be before code and docstring, we skip this ignore "
                "comment. The file will be lint as if the file were not ignored."
            )
        else:
            all_ignored_files.extend(ignored_files)

    not_ignored_issues = get_not_ignore_issue(
        issues,
        all_ignored_lines,
        all_ignored_files,
        config.disabled_rules,
    )

    if warning:
        print()
    for issue in not_ignored_issues:
        CONSOLE.print(str(issue))

    CONSOLE.print(f"Found [bold red]{len(not_ignored_issues)}[/bold red] errors")


@APP.callback()
def main(
    _version: bool = typer.Option(
        False,
        "--version",
        "-v",
        help="Display version of Py-printlinter.",
        callback=version_callback,
        is_eager=True,
    ),
) -> None:
    """
    Display the version if we use the option --version or -v.

    Args:
        version: Boolean option to know if we want display the version or not.
            Default: False
    """
    return


if __name__ == "__main__":
    main()  # pragma: no cover



"""Issues classes used in the project."""

# Standard imports
from collections import namedtuple
from dataclasses import dataclass
from enum import Enum
from pathlib import Path

Issue = namedtuple("Issue", ["err_code", "name"])


class IssueEnum(Issue, Enum):
    """Issue Enumerable to store issues."""

    PRINTDETECT = Issue(err_code="PPL001", name="print-detected")
    PRETTYPRINTDETECT = Issue(err_code="PPL002", name="prettyprint-detected")
    SYSSTDOUTWRITEDETECT = Issue(err_code="PPL003", name="sys.stdout.write-detected")
    SYSSTDERRWRITEDETECT = Issue(err_code="PPL004", name="sys.stderr.write-detected")
    SYSSTDOUTWRITELINESDETECT = Issue(
        err_code="PPL005", name="sys.stdout.writelines-detected"
    )
    SYSSTDERRWRITELINESDETECT = Issue(
        err_code="PPL006", name="sys.stderr.writelines-detected"
    )


@dataclass
class IssueInfo:
    """Issue information class."""

    issue: IssueEnum
    "Issue type."

    num_line: int
    "Line number of the issue."

    num_col: int
    "Column number of the issue."

    line_as_str: str | None
    "String representation of the line."

    from_file: Path | None
    "File who contain the line with the issue."

    ignore: bool
    "If this issue is ignore or not."

    def __str__(self) -> str:
        """
        Get issue as string.

        Returns:
            The string representation of the issue.
        """
        return (
            f"[bold]{self.from_file}[/bold]:{self.num_line}:{self.num_col}: "
            f"[bold red]{self.issue.err_code}[/bold red] [bold]`[/bold]"
            f"{self.line_as_str}[bold]`[/bold] {self.issue.name}"
        )


@dataclass
class IgnoreLine:
    """Ignore line class."""

    line_num: int
    "Line number of the ignored line."

    error_code: str
    "Code of the ignore issue."

    from_file: Path | None
    "File who contains the ignored line."


@dataclass
class IgnoreFile:
    """Ignore file class."""

    error_code: str
    "Code of the ignore issue."

    from_file: Path | None
    "File who contains the ignored line."



"""Processing ingoed lines functions."""

# Local imports
from .classes import IgnoreFile, IgnoreLine, IssueInfo

FAMILIRY_ERR_CODE = [
    "PPL000",
    "PPL100",
    "PPL200",
    "PPL300",
    "PPL400",
    "PPL500",
    "PPL600",
    "PPL700",
    "PPL800",
    "PPL900",
]


def get_not_ignore_issue(
    issues: list[IssueInfo],
    ignore_lines: list[IgnoreLine],
    ignore_files: list[IgnoreFile],
    disabled_rules: list[str],
) -> list[IssueInfo]:
    """
    Get all not ignored issues.

    Args:
        issues: All issue found.
        ignore_lines: All ignored lines.
        ignore_files: All ignored files.
        disabled_rules: All disabled rules.

    Returns:
        All lines are not ignored.
    """
    result = []

    for issue in issues:
        if issue.issue.err_code in disabled_rules:
            continue

        ignore_all_file_equivalence = IgnoreFile(
            error_code="ALL",
            from_file=issue.from_file,
        )
        if ignore_all_file_equivalence in ignore_files:
            continue

        # Ignore a file with a family of code (like PPL000, PPL100, ...)
        for index in range(len(FAMILIRY_ERR_CODE) - 1):
            err_code_min = FAMILIRY_ERR_CODE[index]
            err_code_max = FAMILIRY_ERR_CODE[index + 1]
            if err_code_min < issue.issue.err_code < err_code_max:
                ignored_err_code = err_code_min

        ignore_family_file_equivalence = IgnoreFile(
            error_code=ignored_err_code,
            from_file=issue.from_file,
        )
        if ignore_family_file_equivalence in ignore_files:
            continue

        # Files with specific code (like PPL001, PPL005, ...)
        ignore_file_equivalence = IgnoreFile(
            error_code=issue.issue.err_code,
            from_file=issue.from_file,
        )
        if ignore_file_equivalence in ignore_files:
            continue

        # Ignored lines
        ignore_line_equivalence = IgnoreLine(
            issue.num_line,
            error_code=issue.issue.err_code,
            from_file=issue.from_file,
        )
        if ignore_line_equivalence in ignore_lines:
            continue

        result.append(issue)

    return result



# Local imports
"""Print-Linter project's code."""

# Local imports
from .classes import IgnoreFile, IgnoreLine, IssueEnum, IssueInfo  # noqa: F401
from .config import Config  # noqa: F401
from .ignored_processing import get_not_ignore_issue  # noqa: F401
from .parser import (  # noqa: F401
    enumerate_file,
    get_ignore_files,
    get_ignore_lines,
    parse_file,
)
from .visitor import PrintNodeVisitor, contains_print  # noqa: F401

__version__ = "0.1.0"
__app_name__ = "py-printlinter"
